import typing

from ha_services.mqtt4homeassistant.components import BaseComponent
from ha_services.mqtt4homeassistant.data_classes import ComponentConfig, ComponentState, StatePayload


if typing.TYPE_CHECKING:
    from ha_services.mqtt4homeassistant.device import MqttDevice


class Sensor(BaseComponent):
    """
    MQTT Sensor component for Home Assistant.
    https://www.home-assistant.io/integrations/sensor.mqtt/
    """

    def __init__(
        self,
        *,
        device: 'MqttDevice',
        name: str,
        uid: str,
        component: str = 'sensor',
        #
        # https://www.home-assistant.io/integrations/sensor/#device-class
        device_class: str | None = None,  # e.g.: 'temperature'
        #
        # https://developers.home-assistant.io/docs/core/entity/sensor/#available-state-classes
        state_class: str | None = None,  # e.g.: 'measurement'
        unit_of_measurement: str | None = None,  # e.g.: '°C' / 'W' etc.
        suggested_display_precision: int | None = None,
    ):
        super().__init__(device=device, name=name, uid=uid, component=component)

        self.device_class = device_class
        self.state_class = state_class
        self.unit_of_measurement = unit_of_measurement
        self.suggested_display_precision = suggested_display_precision

        self.value = None  # set_state() must be called to set the value

    def set_state(self, state: StatePayload):
        self.value = state

    def get_state(self) -> ComponentState:
        # e.g.: {'topic': 'homeassistant/sensor/My-device/Chip-Temperature/state', 'payload': '40'}
        return ComponentState(
            topic=f'{self.topic_prefix}/state',
            payload=self.value,
        )

    def get_config(self) -> ComponentConfig:
        payload = {
            'component': self.component,
            'device': self.device.get_mqtt_payload(),
            'device_class': self.device_class,  # e.g.: 'temperature'
            'name': self.name,
            'unique_id': self.uid,
            'unit_of_measurement': self.unit_of_measurement,
            'state_class': self.state_class,  # e.g.: 'measurement'
            'state_topic': f'{self.topic_prefix}/state',
            'json_attributes_topic': f'{self.topic_prefix}/attributes',
        }
        if self.suggested_display_precision is not None:
            payload['suggested_display_precision'] = self.suggested_display_precision
        return ComponentConfig(topic=f'{self.topic_prefix}/config', payload=payload)



#!/usr/bin/env python3

"""
    bootstrap CLI
    ~~~~~~~~~~~~~

    Just call this file, and the magic happens ;)
"""

import hashlib
import shlex
import subprocess
import sys
import venv
from pathlib import Path


def print_no_pip_error():
    print('Error: Pip not available!')
    print('Hint: "apt-get install python3-venv"\n')


try:
    from ensurepip import version
except ModuleNotFoundError as err:
    print(err)
    print('-' * 100)
    print_no_pip_error()
    raise
else:
    if not version():
        print_no_pip_error()
        sys.exit(-1)


assert sys.version_info >= (3, 10), f'Python version {sys.version_info} is too old!'


if sys.platform == 'win32':  # wtf
    # Files under Windows, e.g.: .../.venv/Scripts/python.exe
    BIN_NAME = 'Scripts'
    FILE_EXT = '.exe'
else:
    # Files under Linux/Mac and all other than Windows, e.g.: .../.venv/bin/python
    BIN_NAME = 'bin'
    FILE_EXT = ''

BASE_PATH = Path(__file__).parent
VENV_PATH = BASE_PATH / '.venv'
BIN_PATH = VENV_PATH / BIN_NAME
PYTHON_PATH = BIN_PATH / f'python{FILE_EXT}'
PIP_PATH = BIN_PATH / f'pip{FILE_EXT}'
PIP_SYNC_PATH = BIN_PATH / f'pip-sync{FILE_EXT}'

DEP_LOCK_PATH = BASE_PATH / 'requirements.dev.txt'
DEP_HASH_PATH = VENV_PATH / '.dep_hash'

# script file defined in pyproject.toml as [console_scripts]
# (Under Windows: ".exe" not added!)
PROJECT_SHELL_SCRIPT = BIN_PATH / 'ha_services_dev'


def get_dep_hash():
    """Get SHA512 hash from lock file content."""
    return hashlib.sha512(DEP_LOCK_PATH.read_bytes()).hexdigest()


def store_dep_hash():
    """Generate .venv/.dep_hash"""
    DEP_HASH_PATH.write_text(get_dep_hash())


def venv_up2date():
    """Is existing .venv is up-to-date?"""
    if DEP_HASH_PATH.is_file():
        return DEP_HASH_PATH.read_text() == get_dep_hash()
    return False


def verbose_check_call(*popen_args):
    print(f'\n+ {shlex.join(str(arg) for arg in popen_args)}\n')
    return subprocess.check_call(popen_args)


def main(argv):
    assert DEP_LOCK_PATH.is_file(), f'File not found: "{DEP_LOCK_PATH}" !'

    # Create virtual env in ".venv/":
    if not PYTHON_PATH.is_file():
        print('Create virtual env here:', VENV_PATH.absolute())
        builder = venv.EnvBuilder(symlinks=True, upgrade=True, with_pip=True)
        builder.create(env_dir=VENV_PATH)
        # Update pip
        verbose_check_call(PYTHON_PATH, '-m', 'pip', 'install', '-U', 'pip')

    if not PIP_SYNC_PATH.is_file():
        # Install pip-tools
        verbose_check_call(PYTHON_PATH, '-m', 'pip', 'install', '-U', 'pip-tools')

    if not PROJECT_SHELL_SCRIPT.is_file() or not venv_up2date():
        # install requirements via "pip-sync"
        verbose_check_call(PIP_SYNC_PATH, str(DEP_LOCK_PATH))

        # install project
        verbose_check_call(PIP_PATH, 'install', '--no-deps', '-e', '.')
        store_dep_hash()

    # Call our entry point CLI:
    try:
        verbose_check_call(PROJECT_SHELL_SCRIPT, *argv[1:])
    except subprocess.CalledProcessError as err:
        sys.exit(err.returncode)


if __name__ == '__main__':
    main(sys.argv)



import dataclasses
import socket
from unittest import TestCase

from ha_services.mqtt4homeassistant.data_classes import MqttSettings
from ha_services.mqtt4homeassistant.utilities.string_utils import slugify


class DataClassesTestCase(TestCase):
    def test_mqtt_settings(self):
        with self.assertLogs():
            mqtt_settings = MqttSettings()
        self.assertEqual(
            dataclasses.asdict(mqtt_settings),
            {
                'host': 'mqtt.eclipseprojects.io',
                'port': 1883,
                'password': '',
                'user_name': '',
                'main_uid': slugify(socket.gethostname(), sep='_'),
            },
        )



from cli_base.cli_tools.test_utils.assertion import assert_in
from cli_base.cli_tools.test_utils.cli_readme import AssertCliHelpInReadme
from cli_base.cli_tools.test_utils.rich_test_utils import (
    assert_no_color_env,
    assert_rich_click_no_color,
    assert_rich_no_color,
    assert_subprocess_rich_diagnose_no_color,
)
from cli_base.toml_settings.test_utils.cli_mock import TomlSettingsCliMock
from manageprojects.tests.base import BaseTestCase

from ha_services import constants
from ha_services.cli_app.settings import SETTINGS_DIR_NAME, SETTINGS_FILE_NAME
from ha_services.cli_dev import PACKAGE_ROOT
from ha_services.example import DemoSettings


TERM_WIDTH = 100


class ReadmeTestCase(BaseTestCase):
    cli_mock = None

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()

        settings_overwrites = dict(
            systemd=dict(
                template_context=dict(
                    user='MockedUserName',
                    group='MockedUserName',
                )
            ),
        )

        cls.cli_mock = TomlSettingsCliMock(
            SettingsDataclass=DemoSettings,
            settings_overwrites=settings_overwrites,
            dir_name=SETTINGS_DIR_NAME,
            file_name=SETTINGS_FILE_NAME,
            width=TERM_WIDTH,
        )
        cls.cli_mock.__enter__()

        cls.readme_assert = AssertCliHelpInReadme(base_path=PACKAGE_ROOT, cli_epilog=constants.CLI_EPILOG)

    @classmethod
    def tearDownClass(cls) -> None:
        super().tearDownClass()
        cls.cli_mock.__exit__(None, None, None)

    def test_cli_mock(self):
        assert_no_color_env(width=TERM_WIDTH)
        assert_subprocess_rich_diagnose_no_color(width=TERM_WIDTH)
        assert_rich_no_color(width=TERM_WIDTH)
        assert_rich_click_no_color(width=TERM_WIDTH)

    def invoke_cli(self, *args):
        return self.cli_mock.invoke(cli_bin=PACKAGE_ROOT / 'cli.py', args=args, strip_line_prefix='Usage: ')

    def invoke_dev_cli(self, *args):
        return self.cli_mock.invoke(cli_bin=PACKAGE_ROOT / 'dev-cli.py', args=args, strip_line_prefix='Usage: ')

    def test_main_help(self):
        stdout = self.invoke_cli('--help')
        assert_in(
            content=stdout,
            parts=(
                'Usage: ./cli.py [OPTIONS] COMMAND [ARGS]...',
                'edit-settings',
                'publish-loop',
                constants.CLI_EPILOG,
            ),
        )
        self.readme_assert.assert_block(text_block=stdout, marker='app help')

    def test_dev_help(self):
        stdout = self.invoke_dev_cli('--help')
        assert_in(
            content=stdout,
            parts=(
                'Usage: ./dev-cli.py [OPTIONS] COMMAND [ARGS]...',
                'fix-code-style',
                'tox',
                constants.CLI_EPILOG,
            ),
        )
        self.readme_assert.assert_block(text_block=stdout, marker='dev help')



import datetime
from unittest import TestCase

from ha_services.mqtt4homeassistant.utilities.system_utils import get_system_start_datetime, process_start_datetime


class SystemUtilsTestCase(TestCase):

    def test_get_system_start_datetime(self):
        start_dt = get_system_start_datetime()
        self.assertIsInstance(start_dt, datetime.datetime)

    def test_process_start_datetime(self):
        start_dt = process_start_datetime()
        self.assertIsInstance(start_dt, datetime.datetime)



from unittest import TestCase

from frozendict import frozendict

from ha_services.mqtt4homeassistant.components.switch import Switch
from ha_services.mqtt4homeassistant.device import MainMqttDevice, MqttDevice


class DeviceTestCase(TestCase):
    def test_device(self):
        device = MqttDevice(name='My Device', uid='device1')
        self.assertEqual(
            device.get_mqtt_payload(),
            {'name': 'My Device', 'identifiers': 'device1'},
        )
        self.assertEqual(device.topic_prefix, 'homeassistant')
        self.assertEqual(device.uid, 'device1')

        switch1 = Switch(device=device, name='My Switch 1', uid='switch1')
        switch2 = Switch(device=device, name='My Switch 2', uid='switch2')

        with self.assertRaises(AssertionError) as cm:
            Switch(device=device, name='My Switch 1', uid='switch1')
        self.assertEqual(str(cm.exception), 'Duplicate component: device1-switch1')

        self.assertEqual(
            device.components,
            {
                'device1-switch1': switch1,
                'device1-switch2': switch2,
            },
        )

    def test_device_with_main_device(self):
        main_device = MainMqttDevice(name='Main Device', uid='main_device')
        device = MqttDevice(main_device=main_device, name='Sub Device', uid='sub_device')
        self.assertEqual(
            device.get_mqtt_payload(),
            {
                'name': 'Sub Device',
                'identifiers': 'main_device-sub_device',
                'via_device': 'main_device',
            },
        )
        self.assertEqual(device.topic_prefix, 'homeassistant')
        self.assertEqual(device.uid, 'main_device-sub_device')

    def test_device_extras(self):
        device = MqttDevice(
            name='My Device',
            uid='device1',
            manufacturer='Tinkerforge',
            model='HAT Zero Brick',
            sw_version='1.2.3',
        )
        mqtt_payload = device.get_mqtt_payload()
        self.assertEqual(
            mqtt_payload,
            {
                'name': 'My Device',
                'identifiers': 'device1',
                'manufacturer': 'Tinkerforge',
                'model': 'HAT Zero Brick',
                'sw_version': '1.2.3',
            },
        )
        self.assertIsInstance(mqtt_payload, frozendict)



import logging

import rich_click as click
from cli_base.cli_tools.verbosity import OPTION_KWARGS_VERBOSE, setup_logging
from cli_base.systemd.api import ServiceControl
from rich import print  # noqa

from ha_services.cli_app import cli
from ha_services.cli_app.settings import get_user_settings
from ha_services.example import DemoSettings, SystemdServiceInfo


logger = logging.getLogger(__name__)


@cli.command()
@click.option('-v', '--verbosity', **OPTION_KWARGS_VERBOSE)
def systemd_debug(verbosity: int):
    """
    Print Systemd service template + context + rendered file content.
    """
    setup_logging(verbosity=verbosity)
    user_settings: DemoSettings = get_user_settings(debug=True)
    systemd_settings: SystemdServiceInfo = user_settings.systemd

    ServiceControl(info=systemd_settings).debug_systemd_config()


@cli.command()
@click.option('-v', '--verbosity', **OPTION_KWARGS_VERBOSE)
def systemd_setup(verbosity: int):
    """
    Write Systemd service file, enable it and (re-)start the service. (May need sudo)
    """
    setup_logging(verbosity=verbosity)
    user_settings: DemoSettings = get_user_settings(debug=True)
    systemd_settings: SystemdServiceInfo = user_settings.systemd

    ServiceControl(info=systemd_settings).setup_and_restart_systemd_service()


@cli.command()
@click.option('-v', '--verbosity', **OPTION_KWARGS_VERBOSE)
def systemd_remove(verbosity: int):
    """
    Write Systemd service file, enable it and (re-)start the service. (May need sudo)
    """
    setup_logging(verbosity=verbosity)
    user_settings: DemoSettings = get_user_settings(debug=True)
    systemd_settings: SystemdServiceInfo = user_settings.systemd

    ServiceControl(info=systemd_settings).remove_systemd_service()


@cli.command()
@click.option('-v', '--verbosity', **OPTION_KWARGS_VERBOSE)
def systemd_status(verbosity: int):
    """
    Display status of systemd service. (May need sudo)
    """
    setup_logging(verbosity=verbosity)
    user_settings: DemoSettings = get_user_settings(debug=True)
    systemd_settings: SystemdServiceInfo = user_settings.systemd

    ServiceControl(info=systemd_settings).status()


@cli.command()
@click.option('-v', '--verbosity', **OPTION_KWARGS_VERBOSE)
def systemd_stop(verbosity: int):
    """
    Stops the systemd service. (May need sudo)
    """
    setup_logging(verbosity=verbosity)
    user_settings: DemoSettings = get_user_settings(debug=True)
    systemd_settings: SystemdServiceInfo = user_settings.systemd

    ServiceControl(info=systemd_settings).stop()



import re
import unicodedata


def slugify(text, sep=''):
    """
    >>> slugify('Foo & Bar !', sep='_')
    'Foo_Bar'
    >>> slugify('device_id', sep='_')
    'device_id'
    """
    value = unicodedata.normalize('NFKD', text).encode('ascii', 'ignore').decode('ascii')
    value = re.sub(r'[^\w\s-]', sep, value)
    return re.sub(r'[-_\s]+', sep, value).strip(sep)



import abc
import json
import logging
import typing
from functools import cache

from frozendict import frozendict
from paho.mqtt.client import Client, MQTTMessageInfo

import ha_services
from ha_services.mqtt4homeassistant.data_classes import ComponentConfig, ComponentState
from ha_services.mqtt4homeassistant.utilities.assertments import assert_uid


if typing.TYPE_CHECKING:
    from ha_services.mqtt4homeassistant.device import MqttDevice


logger = logging.getLogger(__name__)


@cache
def get_origin_data() -> dict:
    return {
        'name': 'ha-services',
        'sw_version': ha_services.__version__,
        'support_url': 'https://pypi.org/project/ha_services/',
    }


class BaseComponent(abc.ABC):
    def __init__(
        self,
        *,
        device: 'MqttDevice',
        name: str,
        uid: str,
        component: str,
    ):
        self.device = device

        self.name = name

        assert_uid(uid)
        self.uid = f'{self.device.uid}-{uid}'

        assert component
        self.component = component
        self.device.register_component(component=self)

        # e.g.: 'homeassistant/sensor/My-device/Chip-Temperature'
        self.topic_prefix = f'{self.device.topic_prefix}/{self.component}/{self.device.uid}/{self.uid}'

        self._config_kwargs_cache = None

    def _get_config_kwargs(self) -> dict:
        if self._config_kwargs_cache is None:
            config: ComponentConfig = self.get_config()
            payload = config.payload
            payload['origin'] = get_origin_data()
            payload = json.dumps(config.payload, ensure_ascii=False, sort_keys=True)
            self._config_kwargs_cache = frozendict(
                {
                    'topic': config.topic,
                    'payload': payload,
                }
            )
        return self._config_kwargs_cache

    def publish_config(self, client: Client) -> MQTTMessageInfo:
        config_kwargs = self._get_config_kwargs()
        logger.debug(f'Publishing {self.uid=} config: {config_kwargs}')
        info: MQTTMessageInfo = client.publish(**config_kwargs)
        return info

    def publish_state(self, client: Client) -> MQTTMessageInfo:
        state: ComponentState = self.get_state()
        info: MQTTMessageInfo = client.publish(topic=state.topic, payload=state.payload)
        return info

    def publish_config_and_state(self, client: Client) -> tuple[MQTTMessageInfo, MQTTMessageInfo]:
        config_info = self.publish_config(client)
        state_info = self.publish_state(client)
        return config_info, state_info

    @abc.abstractmethod
    def get_state(self) -> ComponentState:
        pass

    @abc.abstractmethod
    def get_config(self) -> ComponentConfig:
        pass

    def __str__(self):
        return f'{self.__class__.__name__}({self.uid=})'

    def __repr__(self):
        return f'<{self} {id(self)}>'



import subprocess
from unittest import TestCase

from bx_py_utils.path import assert_is_file
from cli_base.cli_tools.code_style import assert_code_style
from manageprojects.test_utils.project_setup import check_editor_config, get_py_max_line_length
from packaging.version import Version

from ha_services import __version__
from ha_services.cli_dev import PACKAGE_ROOT


class ProjectSetupTestCase(TestCase):
    def test_version(self):
        self.assertIsNotNone(__version__)

        version = Version(__version__)  # Will raise InvalidVersion() if wrong formatted
        self.assertEqual(str(version), __version__)

        cli_bin = PACKAGE_ROOT / 'cli.py'
        assert_is_file(cli_bin)

        output = subprocess.check_output([cli_bin, 'version'], text=True)
        self.assertIn(f'ha_services v{__version__}', output)

        dev_cli_bin = PACKAGE_ROOT / 'dev-cli.py'
        assert_is_file(dev_cli_bin)

        output = subprocess.check_output([dev_cli_bin, 'version'], text=True)
        self.assertIn(f'ha_services v{__version__}', output)

    def test_code_style(self):
        return_code = assert_code_style(package_root=PACKAGE_ROOT)
        self.assertEqual(return_code, 0, 'Code style error, see output above!')

    def test_check_editor_config(self):
        check_editor_config(package_root=PACKAGE_ROOT)

        max_line_length = get_py_max_line_length(package_root=PACKAGE_ROOT)
        self.assertEqual(max_line_length, 119)



import logging

import rich_click as click
from cli_base.cli_tools.verbosity import OPTION_KWARGS_VERBOSE, setup_logging
from cli_base.toml_settings.api import TomlSettings
from rich import print  # noqa

from ha_services.cli_app import cli
from ha_services.example import DemoSettings


logger = logging.getLogger(__name__)


SETTINGS_DIR_NAME = 'ha-services'
SETTINGS_FILE_NAME = 'ha-services-demo'


def get_toml_settings() -> TomlSettings:
    return TomlSettings(
        dir_name=SETTINGS_DIR_NAME,
        file_name=SETTINGS_FILE_NAME,
        settings_dataclass=DemoSettings(),
    )


def get_user_settings(debug=True) -> DemoSettings:
    toml_settings: TomlSettings = get_toml_settings()
    user_settings: DemoSettings = toml_settings.get_user_settings(debug=True)
    return user_settings


@cli.command()
@click.option('-v', '--verbosity', **OPTION_KWARGS_VERBOSE)
def edit_settings(verbosity: int):
    """
    Edit the settings file. On first call: Create the default one.
    """
    setup_logging(verbosity=verbosity)
    toml_settings: TomlSettings = get_toml_settings()
    toml_settings.open_in_editor()


@cli.command()
@click.option('-v', '--verbosity', **OPTION_KWARGS_VERBOSE)
def print_settings(verbosity: int):
    """
    Display (anonymized) MQTT server username and password
    """
    setup_logging(verbosity=verbosity)
    toml_settings: TomlSettings = get_toml_settings()
    toml_settings.print_settings()



from ha_services.mqtt4homeassistant.utilities.string_utils import slugify


def assert_uid(uid: str):
    assert uid
    slug = slugify(uid, sep='_')
    assert uid == slug, f'Invalid: {uid=} (slugify: {slug!r})'



import sys
from pathlib import Path

from cli_base.cli_tools.dev_tools import run_unittest_cli
from cli_base.cli_tools.subprocess_utils import verbose_check_call
from manageprojects.utilities.publish import publish_package

import ha_services
from ha_services.cli_dev import PACKAGE_ROOT, cli


@cli.command()
def install():
    """
    Run pip-sync and install 'ha_services' via pip as editable.
    """
    verbose_check_call('pip-sync', PACKAGE_ROOT / 'requirements.dev.txt')
    verbose_check_call('pip', 'install', '--no-deps', '-e', '.')


@cli.command()
def safety():
    """
    Run safety check against current requirements files
    """
    verbose_check_call('safety', 'check', '-r', 'requirements.dev.txt')


@cli.command()
def update():
    """
    Update "requirements*.txt" dependencies files
    """
    bin_path = Path(sys.executable).parent

    verbose_check_call(bin_path / 'pip', 'install', '-U', 'pip')
    verbose_check_call(bin_path / 'pip', 'install', '-U', 'pip-tools')

    extra_env = dict(
        CUSTOM_COMPILE_COMMAND='./dev-cli.py update',
    )

    pip_compile_base = [
        bin_path / 'pip-compile',
        '--verbose',
        '--allow-unsafe',  # https://pip-tools.readthedocs.io/en/latest/#deprecations
        '--resolver=backtracking',  # https://pip-tools.readthedocs.io/en/latest/#deprecations
        '--upgrade',
        '--generate-hashes',
    ]

    # Only "prod" dependencies:
    verbose_check_call(
        *pip_compile_base,
        'pyproject.toml',
        '--output-file',
        'requirements.txt',
        extra_env=extra_env,
    )

    # dependencies + "dev"-optional-dependencies:
    verbose_check_call(
        *pip_compile_base,
        'pyproject.toml',
        '--extra=dev',
        '--output-file',
        'requirements.dev.txt',
        extra_env=extra_env,
    )

    verbose_check_call(bin_path / 'safety', 'check', '-r', 'requirements.dev.txt')

    # Install new dependencies in current .venv:
    verbose_check_call(bin_path / 'pip-sync', 'requirements.dev.txt')


@cli.command()
def publish():
    """
    Build and upload this project to PyPi
    """
    run_unittest_cli(verbose=False, exit_after_run=False)  # Don't publish a broken state

    publish_package(
        module=ha_services,
        package_path=PACKAGE_ROOT,
        distribution_name='ha-services',
    )



"""
    CLI for development
"""

import logging
import sys

import rich_click as click
from bx_py_utils.path import assert_is_file
from cli_base.autodiscover import import_all_files
from cli_base.cli_tools.dev_tools import run_coverage, run_tox, run_unittest_cli
from cli_base.cli_tools.version_info import print_version
from rich.console import Console
from rich.traceback import install as rich_traceback_install
from rich_click import RichGroup
from typeguard import install_import_hook

import ha_services
from ha_services import constants


# Check type annotations via typeguard in all tests.
# Sadly we must activate this here and can't do this in ./tests/__init__.py
install_import_hook(packages=('ha_services',))


logger = logging.getLogger(__name__)


PACKAGE_ROOT = constants.BASE_PATH.parent
assert_is_file(PACKAGE_ROOT / 'pyproject.toml')  # Exists only in cloned git repo


class ClickGroup(RichGroup):  # FIXME: How to set the "info_name" easier?
    def make_context(self, info_name, *args, **kwargs):
        info_name = './dev-cli.py'
        return super().make_context(info_name, *args, **kwargs)


@click.group(
    cls=ClickGroup,
    epilog=constants.CLI_EPILOG,
)
def cli():
    pass


# Register all click commands, just by import all files in this package:
import_all_files(package=__package__, init_file=__file__)


@cli.command()
def version():
    """Print version and exit"""
    # Pseudo command, because the version always printed on every CLI call ;)
    sys.exit(0)


def main():
    print_version(ha_services)

    console = Console()
    rich_traceback_install(
        width=console.size.width,  # full terminal width
        show_locals=True,
        suppress=[click],
        max_frames=2,
    )

    if len(sys.argv) >= 2:
        # Check if we can just pass a command call to origin CLI:
        command = sys.argv[1]
        command_map = {
            'test': run_unittest_cli,
            'tox': run_tox,
            'coverage': run_coverage,
        }
        if real_func := command_map.get(command):
            real_func(argv=sys.argv, exit_after_run=True)

    # Execute Click CLI:
    cli()



import datetime
import os

import psutil


def get_system_start_datetime() -> datetime.datetime:
    start_dt = datetime.datetime.fromtimestamp(psutil.boot_time())
    return start_dt


def process_start_datetime() -> datetime.datetime:
    p = psutil.Process(os.getpid())
    create_time: float = p.create_time()
    start_dt = datetime.datetime.fromtimestamp(create_time, datetime.timezone.utc)
    return start_dt



from unittest import TestCase

from ha_services.mqtt4homeassistant.components.sensor import Sensor
from ha_services.mqtt4homeassistant.data_classes import ComponentConfig, ComponentState
from ha_services.mqtt4homeassistant.device import MqttDevice


class SensorTestCase(TestCase):
    def test_mimimal_sensor(self):
        sensor = Sensor(
            device=MqttDevice(name='My device', uid='device_id'),
            name='My component',
            uid='component_id',
        )
        self.assertEqual(
            sensor.get_config(),
            ComponentConfig(
                topic='homeassistant/sensor/device_id/device_id-component_id/config',
                payload={
                    'component': 'sensor',
                    'device': {'identifiers': 'device_id', 'name': 'My device'},
                    'device_class': None,
                    'json_attributes_topic': 'homeassistant/sensor/device_id/device_id-component_id/attributes',
                    'name': 'My component',
                    'state_class': None,
                    'state_topic': 'homeassistant/sensor/device_id/device_id-component_id/state',
                    'unique_id': 'device_id-component_id',
                    'unit_of_measurement': None,
                },
            ),
        )
        self.assertEqual(
            sensor.get_state(),
            ComponentState(topic='homeassistant/sensor/device_id/device_id-component_id/state', payload=None),
        )
        sensor.set_state(123)
        self.assertEqual(
            sensor.get_state(),
            ComponentState(topic='homeassistant/sensor/device_id/device_id-component_id/state', payload=123),
        )

    def test_full_sensor(self):
        sensor = Sensor(
            device=MqttDevice(name='My device', uid='device_id'),
            name='My component',
            uid='component_id',
            component='sensor',
            device_class='temperature',
            state_class='measurement',
            unit_of_measurement='°C',
            suggested_display_precision=2,
        )
        self.assertEqual(
            sensor.get_config(),
            ComponentConfig(
                topic='homeassistant/sensor/device_id/device_id-component_id/config',
                payload={
                    'component': 'sensor',
                    'device': {'identifiers': 'device_id', 'name': 'My device'},
                    'device_class': 'temperature',
                    'json_attributes_topic': 'homeassistant/sensor/device_id/device_id-component_id/attributes',
                    'name': 'My component',
                    'state_class': 'measurement',
                    'state_topic': 'homeassistant/sensor/device_id/device_id-component_id/state',
                    'suggested_display_precision': 2,
                    'unique_id': 'device_id-component_id',
                    'unit_of_measurement': '°C',
                },
            ),
        )
        self.assertEqual(
            sensor.get_state(),
            ComponentState(topic='homeassistant/sensor/device_id/device_id-component_id/state', payload=None),
        )
        sensor.set_state(456)
        self.assertEqual(
            sensor.get_state(),
            ComponentState(topic='homeassistant/sensor/device_id/device_id-component_id/state', payload=456),
        )



from unittest.mock import patch

from bx_py_utils.test_utils.context_managers import MassContextManager

from ha_services.mqtt4homeassistant.mocks.psutil_mock import PsutilMock


class MqttClientMock:
    def __init__(self):
        self.messages = []

    def publish(self, **kwargs) -> None:
        self.messages.append(kwargs)


class MainMqttDeviceMock(MassContextManager):
    def __init__(self):

        psutil_mock = PsutilMock()

        self.mocks = (
            patch('ha_services.mqtt4homeassistant.device.socket.gethostname', return_value='TheHostName'),
            #
            patch('ha_services.mqtt4homeassistant.system_info.cpu.psutil', psutil_mock),
            patch('ha_services.mqtt4homeassistant.system_info.memory.psutil', psutil_mock),
            patch('ha_services.mqtt4homeassistant.system_info.temperatures.psutil', psutil_mock),
        )



import rich_click as click
from cli_base.cli_tools import code_style
from cli_base.cli_tools.verbosity import OPTION_KWARGS_VERBOSE
from cli_base.click_defaults import OPTION_ARGS_DEFAULT_TRUE

from ha_services.cli_dev import PACKAGE_ROOT, cli


@cli.command()
@click.option('--color/--no-color', **OPTION_ARGS_DEFAULT_TRUE)
@click.option('-v', '--verbosity', **OPTION_KWARGS_VERBOSE)
def fix_code_style(color: bool, verbosity: int):
    """
    Fix code style of all ha_services source code files via darker
    """
    code_style.fix(package_root=PACKAGE_ROOT, darker_color=color, darker_verbose=verbosity > 0)


@cli.command()
@click.option('--color/--no-color', **OPTION_ARGS_DEFAULT_TRUE)
@click.option('-v', '--verbosity', **OPTION_KWARGS_VERBOSE)
def check_code_style(color: bool, verbosity: int):
    """
    Check code style by calling darker + flake8
    """
    code_style.check(package_root=PACKAGE_ROOT, darker_color=color, darker_verbose=verbosity > 0)



from unittest import TestCase

from ha_services.mqtt4homeassistant.components.switch import Switch
from ha_services.mqtt4homeassistant.data_classes import ComponentConfig, ComponentState
from ha_services.mqtt4homeassistant.device import MqttDevice


class SwitchTestCase(TestCase):
    def test_switch(self):
        switch = Switch(
            device=MqttDevice(name='My Device', uid='device1'),
            name='My Switch',
            uid='switch1',
        )
        self.assertEqual(
            switch.get_config(),
            ComponentConfig(
                topic='homeassistant/switch/device1/device1-switch1/config',
                payload={
                    'command_topic': 'homeassistant/switch/device1/device1-switch1/command',
                    'component': 'switch',
                    'device': {'identifiers': 'device1', 'name': 'My Device'},
                    'json_attributes_topic': 'homeassistant/switch/device1/device1-switch1/attributes',
                    'name': 'My Switch',
                    'payload_off': 'OFF',
                    'payload_on': 'ON',
                    'state_topic': 'homeassistant/switch/device1/device1-switch1/state',
                    'unique_id': 'device1-switch1',
                },
            ),
        )
        self.assertEqual(
            switch.get_state(),
            ComponentState(topic='homeassistant/switch/device1/device1-switch1/state', payload=None),
        )
        with self.assertRaises(AssertionError) as cm:
            switch.set_state('invalid_state')
        self.assertEqual(str(cm.exception), 'Invalid state: invalid_state')

        switch.set_state('ON')
        self.assertEqual(
            switch.get_state(),
            ComponentState(topic='homeassistant/switch/device1/device1-switch1/state', payload='ON'),
        )



from bx_py_utils.test_utils.unittest_utils import BaseDocTests

import ha_services


class DocTests(BaseDocTests):
    def test_doctests(self):
        self.run_doctests(
            modules=(ha_services,),
        )



import datetime
import json
from unittest import TestCase
from unittest.mock import patch

import frozendict
from bx_py_utils.test_utils.snapshot import assert_snapshot

from ha_services.mqtt4homeassistant.components.sensor import Sensor
from ha_services.mqtt4homeassistant.data_classes import ComponentConfig, ComponentState
from ha_services.mqtt4homeassistant.device import MainMqttDevice, MqttDevice
from ha_services.mqtt4homeassistant.mocks.mqtt_client_mock import MainMqttDeviceMock, MqttClientMock


class IntergrationTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.get_origin_data_patch = patch(
            'ha_services.mqtt4homeassistant.components.get_origin_data',
            return_value={
                'name': 'ha-services',
                'support_url': 'https://pypi.org/project/ha_services/',
                'sw_version': '1.2.3',
            },
        )
        cls.get_origin_data_patch.__enter__()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        cls.get_origin_data_patch.__exit__(None, None, None)

    def test_main_sub(self):
        main_device = MainMqttDevice(
            name='Main Device',
            uid='main_uid',
        )
        device = MqttDevice(
            main_device=main_device,
            name='Sub Device',
            uid='sub_uid',
        )
        test_sensor = Sensor(
            device=device,
            name='Test Sensor',
            uid='sensor_uid',
        )
        self.assertEqual(
            test_sensor.get_state(),
            ComponentState(
                topic='homeassistant/sensor/main_uid-sub_uid/main_uid-sub_uid-sensor_uid/state',
                payload=None,
            ),
        )
        test_sensor.set_state(123)
        self.assertEqual(
            test_sensor.get_state(),
            ComponentState(
                topic='homeassistant/sensor/main_uid-sub_uid/main_uid-sub_uid-sensor_uid/state',
                payload=123,
            ),
        )

        self.assertEqual(
            test_sensor.get_config(),
            ComponentConfig(
                topic='homeassistant/sensor/main_uid-sub_uid/main_uid-sub_uid-sensor_uid/config',
                payload={
                    'component': 'sensor',
                    'device': frozendict.frozendict(
                        {
                            'name': 'Sub Device',
                            'identifiers': 'main_uid-sub_uid',
                            'via_device': 'main_uid',
                        }
                    ),
                    'device_class': None,
                    'json_attributes_topic': (
                        'homeassistant/sensor/main_uid-sub_uid/main_uid-sub_uid-sensor_uid/attributes'
                    ),
                    'name': 'Test Sensor',
                    'state_class': None,
                    'state_topic': 'homeassistant/sensor/main_uid-sub_uid/main_uid-sub_uid-sensor_uid/state',
                    'unique_id': 'main_uid-sub_uid-sensor_uid',
                    'unit_of_measurement': None,
                },
            ),
        )

        mqtt_client_mock = MqttClientMock()
        with MainMqttDeviceMock(), self.assertLogs('ha_services', level='DEBUG'):
            main_device.poll_and_publish(mqtt_client_mock)

        first_message = mqtt_client_mock.messages[0]
        self.assertEqual(first_message['topic'], 'homeassistant/sensor/main_uid/main_uid-hostname/config')
        self.assertEqual(
            json.loads(first_message['payload']),
            {
                'component': 'sensor',
                'device': {'identifiers': 'main_uid', 'name': 'Main Device'},
                'device_class': None,
                'json_attributes_topic': 'homeassistant/sensor/main_uid/main_uid-hostname/attributes',
                'name': 'Hostname',
                'origin': {
                    'name': 'ha-services',
                    'support_url': 'https://pypi.org/project/ha_services/',
                    'sw_version': '1.2.3',
                },
                'state_class': None,
                'state_topic': 'homeassistant/sensor/main_uid/main_uid-hostname/state',
                'unique_id': 'main_uid-hostname',
                'unit_of_measurement': None,
            },
        )

        replaces = []
        for message in mqtt_client_mock.messages:
            payload = message['payload']
            self.assertIsInstance(payload, (int, float, str), message)

            if message['topic'] == 'homeassistant/sensor/main_uid/main_uid-up_time/state':
                datetime.datetime.fromisoformat(message['payload'])
                message['payload'] = '<mocked up_time>'
                replaces.append('up_time')

            if message['topic'] == 'homeassistant/sensor/main_uid/main_uid-process_start/state':
                datetime.datetime.fromisoformat(message['payload'])
                message['payload'] = '<mocked process_start>'
                replaces.append('process_start')

        self.assertEqual(replaces, ['up_time', 'process_start'])

        assert_snapshot(got=mqtt_client_mock.messages)



from __future__ import annotations

import dataclasses
import logging
import socket
from typing import TypeAlias

from bx_py_utils.anonymize import anonymize

from ha_services.mqtt4homeassistant.utilities.string_utils import slugify


logger = logging.getLogger(__name__)

StatePayload: TypeAlias = str | bytes | bytearray | int | float | None


@dataclasses.dataclass
class ComponentState:
    topic: str
    payload: StatePayload


@dataclasses.dataclass
class ComponentConfig:
    topic: str
    payload: dict


@dataclasses.dataclass
class MqttSettings:
    """
    Credentials to MQTT server that should be used.

    The `main_uid` is used as a second prefix for all MQTT messages, to avoid conflicts.
    """

    host: str = 'mqtt.eclipseprojects.io'  # public test MQTT broker service
    port: int = 1883
    user_name: str = ''
    password: str = ''

    main_uid: str = dataclasses.field(default_factory=socket.gethostname)

    def __post_init__(self):
        assert self.main_uid, 'main_uid must be provided'

        main_uid = slugify(self.main_uid, sep='_')
        if main_uid != self.main_uid:
            self.main_uid = main_uid
            logger.warning('main_uid has been slugified to: %r', self.main_uid)

    def anonymized(self):
        data = dataclasses.asdict(self)
        if self.password:
            data['password'] = anonymize(self.password)
        return data
